import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.StringTokenizer;

/**
 * デモ表示のための基底クラス。
 *
 * JFrame を拡張して、次のような機能を提供する。
 * <ul>
 * <li>GraphPanel や制御用のボタンを配置
 * <li>グラフのデータを jar から読み込む readParam()
 * <li>グラフの枝表示の為の drawEdge()
 * <li>表示速度調節のための sleep()
 * </ul>
 * 
 * @author Nobuyuki Tsuchimura
 * @see GraphPanel
 */
abstract public class GraphApplication extends JFrame implements Runnable {
    /** 入力ファイル名 */
    String file;
    /** 頂点数 */
    protected int n;
    /** グラフの頂点座標 */
    protected int x[], y[];
    /** 頂点間の距離 */
    protected int d[][];

    /** 文字色 */
    final static Color FG_COLOR = Color.WHITE;
    /** 背景色 */
    final static Color BG_COLOR = Color.BLACK;
    /** ボタンの色 */
    final static Color BUTTON_COLOR = Color.GREEN.darker().darker();
    /** 文字 */
    final static Font FONT = new Font(Font.MONOSPACED, Font.BOLD, 16);

    protected volatile Thread thread = null;
    protected boolean threadSuspended = false;
    protected boolean stepPressed = false;
    protected boolean initPressed = false;
    private GraphPanel graphPanel;
    protected JPanel northPanel, ctrlPanel, messagePanel;
    protected JButton playButton, stepButton, initButton;

    /** コンストラクタ */
    public GraphApplication(String title, String file) {
        setTitle(title);
        this.file = file;
        //System.out.println("GraphApplication 1.22");
        addWindowListener(new ClosingListener());
        setSize(800, 600);
        setExtendedState(JFrame.MAXIMIZED_BOTH); // 最大化
        setBackground( BG_COLOR );
        setForeground( FG_COLOR );
        graphPanel  = new GraphPanel(this);
        northPanel  = new JPanel();
        ctrlPanel   = new JPanel();
        messagePanel= new JPanel();

        northPanel.setBackground(BG_COLOR);
        ctrlPanel .setBackground(BG_COLOR);
        messagePanel.setBackground(BG_COLOR);

        setLayout( new BorderLayout() );
        add( graphPanel, BorderLayout.CENTER );
        add( northPanel, BorderLayout.NORTH );

        northPanel.setLayout( new FlowLayout() );
        northPanel.add( ctrlPanel );
        northPanel.add( messagePanel );

        // Play, Step, Init の各ボタンを配置する
        playButton = addCtrlButton( "",     new PlayButtonListener() );
        stepButton = addCtrlButton( "Step", new StepButtonListener() );
        initButton = addCtrlButton( "Init", new InitButtonListener() );
        setSuspended(false);
        setReady(false);
    }

    /** 制御用ボタンを配置する。 */
    protected JButton addCtrlButton(String title, ActionListener listener) {
        JButton b = new JButton(title);
        b.setFont(FONT);

//        Mac で背景色が変わらないので、すべてディフォルトに
//        http://stackoverflow.com/questions/1065691/how-to-set-the-background-color-of-a-jbutton-on-the-mac-os
        //b.setBackground(BUTTON_COLOR);
        //b.setForeground(FG_COLOR);
        //b.setOpaque(true);

        b.addActionListener(listener);
        ctrlPanel.add(b);
        return b;
    }

    /**
       枝を(アルゴリズムとは関係なく)全て描画する処理を自前で書く。
       <tt>drawEdge(i,j,g)</tt> を呼び出すことで実現する。
       @see #drawEdge(int,int,Graphics)
    */
    abstract public void drawAllEdges( Graphics g );

    /**
       枝を描画する。drawAllEdges 経由でのみ使ってよい。
       @param i 枝の始点の頂点番号
       @param j 枝の終点の頂点番号
       @param g <tt>drawAllEdges()</tt> で与えられた <tt>Graphics</tt>
       @see #drawAllEdges
       @see #drawEdge(int,int)
    */
    public void drawEdge( int i, int j, Graphics g ) {
        graphPanel.drawEdge( i, j, g );
    }

    /**
       アルゴリズムの途中で追加/削除した枝を描画する。
       @param i 枝の始点の頂点番号
       @param j 枝の終点の頂点番号
       @see #drawEdge(int,int,Graphics)
    */
    public void drawEdge( int i, int j ) {
        graphPanel.drawEdge( i, j );
    }

    /**
       アルゴリズムの途中で強調表示したい頂点を描画する。
       @param i 頂点番号
       @see #drawEdge(int,int,Graphics)
    */
    public void drawNodeCircle( int i ) {
        graphPanel.drawNodeCircle( i );
    }

    public void setColors( Color node, Color edge, Color edgeBg ) {
        graphPanel.setColors( node, edge, edgeBg );
    }

    /**
     * Jarファイル内のファイルを読むためのBufferedReaderを取得する。
     * http://d.hatena.ne.jp/kirifue/20090610/1245215935
     * @param path
     * @return 取得した BufferedReader
     */
    BufferedReader getJarReader(String path){
        BufferedReader reader = null;
        InputStream stream = this.getClass().getClassLoader()
        .getResourceAsStream(path);
        if(stream!=null){
            reader = new BufferedReader(new InputStreamReader(stream));
        }
        return reader;
    }

    /** 
     * {@link #getJarReader(String)} の代替品で、ファイルから読み込む。
     */
    BufferedReader getFileReader(String path){
        try {
            return new BufferedReader(new FileReader(path));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            System.exit(1);
        }
        return null;
    }
    
    /** TSPlib のデータを x[], y[], d[][] に読み込む 
     * @throws IOException 
     * @throws NumberFormatException */
    public void readParam() throws NumberFormatException, IOException {
        boolean att = false;

        BufferedReader br = getJarReader(file);
//        BufferedReader br = getFileReader(file);
        String line;
        while ((line = br.readLine()) != null) {
            StringTokenizer t = new StringTokenizer(line, " :");
            String s = t.nextToken();
            if (s.equals("DIMENSION")) {
                n = Integer.parseInt(t.nextToken());
                x = new int[n];
                y = new int[n];
            } else if (s.equals("EDGE_WEIGHT_TYPE")) {
                att = t.nextToken().equals("ATT");
                //System.out.println("att = " + att);
            } else if (s.equals("EOF")) {
                break;
            } else if (Character.isDigit(s.charAt(0))) {
                int i = Integer.parseInt(s) - 1;  // 1 origin -> 0 origin
                x[i] = (int)Math.round(Double.parseDouble(t.nextToken()));
                y[i] = (int)Math.round(Double.parseDouble(t.nextToken()));
            }
        }
        br.close();
        d = new int[n][n];
        for (int i=0; i<n; i++) {
            for (int j=0; j<i; j++) {
                final int dx = x[i] - x[j];
                final int dy = y[i] - y[j];
                d[i][j] = d[j][i] = att ?
                        (int)Math.ceil( Math.sqrt((dx*dx+dy*dy)/10.0) ) : // att
                        (int)Math.round( Math.sqrt(dx*dx+dy*dy) );       // Euclid2
            }
        }
    }

    /** Pause のサイズのほうが Play より大きいので、
     * この時のサイズを setPreferredSize したら true */
    private boolean pauseSizeSet = false;
    /**
       <tt>sleep()</tt> の呼び出しで停止させるかどうかを設定する。
       @param enable true なら(ボタン操作があるまで)停止、
       false なら sleep 動作させる。
       @see #sleep(int)
    */
    public void setSuspended(boolean enable) {
        if (!pauseSizeSet && enable) {
            pauseSizeSet = true;
            playButton.setPreferredSize(playButton.getSize()); 
        }
        playButton.setText(enable ? "Play" : "Pause");
        playButton.setEnabled(true);
        stepButton.setEnabled(enable);
        threadSuspended = enable;
    }

    /**
     *  制御用ボタンを有効にする。
     * run() から呼び出す。
     */
    public void setReady(boolean enable) {
        playButton.setEnabled(enable);
        stepButton.setEnabled(enable);
        initButton.setEnabled(enable);
    }
    
    /** start() 経由で呼び出される */
    public void run() {
        try {
            readParam();
        } catch (NumberFormatException e) {
            e.printStackTrace();
            return;
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        setSleep( 2000 / (int)Math.pow(n, 0.7) );
        graphPanel.setNodes( x, y );
        repaint();
        while (!isVisible()) {
            System.out.println("Sleep untill visible");
            sleep(300);
        }
    }

    /** 描くべきグラフのデータ構造に変化があれば呼び出す。
        単なる再表示には super.repaint() を使う。
        TODO:repaint()→repaintGraph(), super.repaint()→repaint()
     */
    //@Override
    public void repaint() {
        graphPanel.updateOffGraphics();
        super.repaint();
    }

    public void start() {
        thread = new Thread( this );
        thread.start();
    }

    public void stop() {
        thread = null;
    }

    /** 累積待ち時間 */
    private int wait = 0;
    /** 累積待ち時間がこの値以上になれば、実際に wait する */
    private static final int WAIT_MIN = 50;
    /** 省略時の待ち時間 */
    private int waitDefault = 100;

    /**
       <tt>sleep()</tt> での待ち時間を設定する。
       @see #sleep()
    */
    public void setSleep( int msec ) {
        waitDefault = msec;
    }

    /**
       時間待ちをする。
       待ち時間は <tt>setSleep()</tt> で設定しておく。
       @return false なら中断, true なら正常
       @see #sleep(int)
       @see #setSleep
     */
    public boolean sleep() {  return sleep( waitDefault );  }

    /**
       時間待ちをする。
       ただし、実際に停止するのは、待ち時間が累積して
       <tt>WAIT_MIN</tt> ミリ秒以上になった時だけ。
       また、指定した待ち時間にかかわらず、
       ボタン操作や <tt>setSuspended(true)</tt> により、
       長時間停止させられることもある。
       @param msec 待ち時間（ミリ秒）
       @return false なら中断, true なら正常       
       @see #sleep
       @see #setSuspended
    */
    public boolean sleep( int msec ) {
	//System.out.println( "sleep(" + msec + ")" );
        wait += msec;
        if ( wait < WAIT_MIN && !stepPressed && !threadSuspended ) return true;
        if ( wait < WAIT_MIN ) wait = WAIT_MIN;
        do {
            if ( initPressed ) {
                initPressed = false;
                threadSuspended = false;
                return false;
            }
            if ( stepPressed ) {
                stepPressed = false;
                wait = 0;
                return true;
            }
            //System.out.println( "wait " + wait + stepPressed );
            Thread t = Thread.currentThread();
            if ( thread != t ) return false;
            graphPanel.repaintIfNeeded();
            try {
                Thread.sleep( wait );
            } catch ( InterruptedException e ) {
                return false;
            }
        } while ( threadSuspended );
        wait = 0;
        return true;
    }
    // 参照 http://java.sun.com/j2se/1.4/ja/docs/ja/guide/misc/threadPrimitiveDeprecation.html


    /**
       マウス操作で頂点を選択する。
       @return 頂点番号
    */
    public int getStartingIndex() {
        graphPanel.setMouseHandle( true );
        setSuspended( true );
        sleep();
        graphPanel.capturing = false;
        return graphPanel.selectedIndex;
    }

    /** Init ボタンが押されるまで待つ。*/
    public void sleepForever() {
        playButton.setEnabled( false );
        stepButton.setEnabled( false );
        threadSuspended = false;
        while ( sleep(300) ) /* empty */;
    }

    /** クリックされるまで待つ。
    public void waitForClick() {
        graphPanel.clicked = false;
        while ( !graphPanel.clicked ) sleep( 300 );
    }
    */

    private class PlayButtonListener implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            setSuspended( !threadSuspended );
        }
    }
    private class StepButtonListener implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            stepPressed = true;
        }
    }
    private class InitButtonListener implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            initPressed = true;
        }
    }

    private class ClosingListener extends WindowAdapter {
        //@Override
        public void windowClosing(WindowEvent arg0) {
            stop();
        }
    }
}
